@@ -38,14 +38,14 @@ Example::
     >>> sample = activation.Sample("Co30Fe70", 10)
     >>> sample.calculate_activation(env, exposure=10, rest_times=[0, 1, 24, 360])
     >>> sample.show_table()
-                                          ----------------- activity (uCi) ------------------
-    isotope  product  reaction  half-life        0 hrs        1 hrs       24 hrs      360 hrs
-    -------- -------- -------- ---------- ------------ ------------ ------------ ------------
-    Co-59    Co-60         act    5.272 y     0.000496     0.000496    0.0004958    0.0004933
-    Co-59    Co-60m+       act     10.5 m        1.664       0.0317          ---          ---
-    -------- -------- -------- ---------- ------------ ------------ ------------ ------------
-                                    total        1.665      0.03221    0.0005084     0.000505
-    -------- -------- -------- ---------- ------------ ------------ ------------ ------------
+                                           ----------------- activity (uCi) ------------------
+    isotope  product   reaction  half-life        0 hrs        1 hrs       24 hrs      360 hrs
+    -------- --------- -------- ---------- ------------ ------------ ------------ ------------
+    Co-59    Co-60          act    5.272 y    0.0004959    0.0004959    0.0004958    0.0004933
+    Co-59    Co-60m+        act     10.5 m        1.664      0.03169          ---          ---
+    -------- --------- -------- ---------- ------------ ------------ ------------ ------------
+                                     total        1.664       0.0322    0.0005084    0.0005049
+    -------- --------- -------- ---------- ------------ ------------ ------------ ------------
 
     >>> print("%.3f"%sample.decay_time(0.001)) # number of hours to reach 1 nCi
     2.053
@@ -179,11 +179,16 @@ class Sample(object):
         # Find the small rest time (probably 0 hr)
         min_rest, To = min(enumerate(self.rest_times), key=lambda x: x[1])
         # Find the activity at that time, and the decay rate
-        data = [(Ia[min_rest], LN2/a.Thalf_hrs) for a, Ia in self.activity.items()]
+        data = [
+            (Ia[min_rest], LN2/a.Thalf_hrs)
+            for a, Ia in self.activity.items()
+            # TODO: not sure why Ia is zero, but it messes up the initial value guess if it is there
+            if Ia[min_rest] > 0.0
+            ]
         # Build functions for total activity at time T - target and its derivative
         # This will be zero when activity is at target
         f = lambda t: sum(Ia*exp(-La*(t-To)) for Ia, La in data) - target
-        df = lambda t: sum(La*Ia*(To-1)*exp(-La*(t-To)) for Ia, La in data)
+        df = lambda t: sum(-La*Ia*exp(-La*(t-To)) for Ia, La in data)
         # Return target time, or 0 if target time is negative
         if f(0) < target:
             return 0
@@ -192,6 +197,7 @@ class Sample(object):
         # dominate at long times, but at short times they will not affect the
         # derivative. Choosing a time that satisfies the longest half-life seems
         # to work well enough.
+        #print("data", data, [])
         initial = max(-log(target/Ia)/La + To for Ia, La in data)
         t, ft = find_root(initial, f, df)
         percent_error = 100*abs(ft)/target
@@ -244,8 +250,8 @@ class Sample(object):
         # Print a dashed separator above and below each column
         header = ["isotope", "product", "reaction", "half-life"] \
                  + ["%g hrs"%vi for vi in self.rest_times]
-        separator = ["-"*8, "-"*8, "-"*8, "-"*10] + ["-"*12]*len(self.rest_times)
-        cformat = "%-8s %-8s %8s %10s " + " ".join(["%12s"]*len(self.rest_times))
+        separator = ["-"*8, "-"*9, "-"*8, "-"*10] + ["-"*12]*len(self.rest_times)
+        cformat = "%-8s %-9s %8s %10s " + " ".join(["%12s"]*len(self.rest_times))
 
         width = sum(len(c)+1 for c in separator[4:]) - 1
         if width < 16:
@@ -284,6 +290,7 @@ def find_root(x, f, df, max=20, tol=1e-10):
     """
     fx = f(x)
     for _ in range(max):
+        #print(f"step {_}: {x=} {fx=} f'={df(x)}")
         if abs(f(x)) < tol:
             break
         x -= fx / df(x)
@@ -359,11 +366,58 @@ COLUMN_NAMES = [
 INT_COLUMNS = [1, 2, 4]
 BOOL_COLUMNS = [13]
 FLOAT_COLUMNS = [6, 11, 14, 15, 16, 17, 19, 20, 21]
+UNITS_TO_HOURS = {'y': 8760, 'd': 24, 'h': 1, 'm': 1/60, 's': 1/3600}
 
 def activity(isotope, mass, env, exposure, rest_times):
     """
-    Compute isotope specific daughter products after the given exposure time and rest period.
+    Compute isotope specific daughter products after the given exposure time and
+    rest period.
+
+    Activations are listed in *isotope.neutron_activation*. Most of the
+    activations (n,g n,p n,a n,2n) define a single step process, where a neutron
+    is absorbed yielding the daughter and some prompt radiation. The daughter
+    itself will decay during exposure, yielding a balance between production and
+    emission. Any exposure beyound about eight halflives will not increase
+    activity for that product.
+
+    Activity for daughter products may undergo further neutron capture, reducing
+    the activity of the daughter product but introducing a grand daughter with
+    its own activity.
+
+    The data tables for activation are only precise to about three significant
+    figures. Any changes to the calculations below this threshold, e.g., due to
+    slightly different mass or abundance, are therefore of little concern.
+
+    Differences in formulas compare to the NCNR activation spreadsheet:
+
+    * Column M: Use ln(2) rather than 0.693
+    * Column N: Use ln(2) rather than 0.693
+    * Column O: Rewrite to use expm1(x) = exp(x) - 1::
+        = L (1 - exp(-M Y43)/(1-(M/N) ) + exp(-N Y43)/((N/M)-1) )
+        = L (1 - N exp(-M Y43)/(N-M) + M exp(-N Y43)/(N-M) )
+        = L/(N-M) ( (N-M) - N exp(-M Y43) + M exp(-N Y43) )
+        = L/(N-M) ( N(1 - exp(-M Y43)) + M (exp(-N Y43) - 1) )
+        = L/(N-M) ( -N expm1(-M Y43) + M expm1(-N) )
+        = L/(N-M) (M expm1(-N Y43) - N expm1(-M Y43))
+    * Column X: Rewrite to use expm1(x) = exp(x) - 1::
+        = W ((abs(U)<1e-10 and abs(V)<1e-10) ? (V-U + (V-U)(V+U)/2) : (exp(-U)-exp(-V)))
+        = W (exp(-U) - exp(-V))
+        = W exp(-V) (exp(V-U) - 1) = W exp(-U) (1 - exp(U-V))
+        = W exp(-V) expm1(V-U) = -W exp(-U) expm1(U-V)
+        = (U > V) ? (W exp(-V) expm1(V-U)) : (-W exp(-U) expm1(U-V))
+
+    Differences in the data tables:
+
+    * AW1462 (W-186 => W-188 2n) t1/2 in hrs is not converting days to hours
+    * AK1495 (Au-198 => Au-199 2n) target should be Au-197
+    * AN1428 (Tm-169 => Tm-171 2n) t1/2 updated to Tm-171 rather than Tm-172
+    * AN1420 (Er-162 => Ho-163 b) t1/2 updated to 4570 y from 10 y
     """
+    # TODO: is the table missing 1-H => 3-H ?
+    # 0nly activations which produce radioactive daughter products are
+    # included. Because 1-H => 2-H (act) is not in the table, is this why
+    # there is no entry for 1-H => 3-H (2n).
+
     result = {}
     if not hasattr(isotope, 'neutron_activation'):
         return result
@@ -381,17 +435,24 @@ def activity(isotope, mass, env, exposure, rest_times):
         # Column G: Half-life
         #    ai.Thalf_str
         # Column H: initial effective cross-section (b)
+        #    env.epithermal_reduction_factor:$Q$1 = 1/env.Cd_ratio:$F$15
         initialXS = ai.thermalXS + env.epithermal_reduction_factor*ai.resonance
         # Column I: reaction
         #    ai.reaction
         # Column J: fast?
         #    ai.fast
         # Column K: effective reaction flux (n/cm^2/s)
+        #    env.fluence:$F$13  env.fast_ratio:$F$17
         flux = env.fluence/env.fast_ratio if ai.fast else env.fluence
         # Column L: root part of activation calculation
-        # Decay correction portion done in column M
-        # The given mass is sample mass * sample fraction * isotope abundance
-        root = flux * initialXS * 1e-24 * mass / isotope.isotope * 1.6278e19
+        #    mass:$F$19
+        #    Decay correction portion done in column M
+        #    The given mass is sample mass * sample fraction * isotope abundance
+        #    The constant converts from Bq to uCi via avogadro's number with
+        #        N_A[atoms] / 3.7e10[Bq/Ci] * 1e6 [uCi/Ci] ~ 1.627605611e19
+        Bq_to_uCi = 1.6276e19
+        #Bq_to_uCi = constants.avogadro_number / 3.7e4
+        root = flux * initialXS * 1e-24 * mass / isotope.isotope * Bq_to_uCi
         # Column M: 0.69/t1/2  (1/h) lambda of produced nuclide
         lam = LN2/ai.Thalf_hrs
         #print(ai.thermalXS, ai.resonance, env.epithermal_reduction_factor)
@@ -422,10 +483,6 @@ def activity(isotope, mass, env, exposure, rest_times):
             activity = root/(parent_lam - lam) * (
                 lam*expm1(-parent_lam*exposure) - parent_lam*expm1(-lam*exposure))
             #print("N", parent_lam, "O", activity)
-            if activity < 0:
-                #frac = lam/parent_lam
-                #print(f"{activity=} {root=} {lam=} {exposure=} {parent_lam=} {exp(-lam*exposure)=} {lam/parent_lam=} {exp(-lam*exposure)/(1-(lam/parent_lam))=} {(frac-expm1(-lam*exposure))/(1-frac)=}")
-                activity = 1e-10
         elif ai.reaction == '2n':
             # Column N: 0.69/t1/2 (1/h) lambda of parent nuclide
             parent_lam = LN2 / ai.Thalf_parent
@@ -466,13 +523,13 @@ def activity(isotope, mass, env, exposure, rest_times):
             V = (env.fluence*effectiveXS*3600*1e-24+lam)*exposure
             # Column W: L/(L-nvs1+nvs2)
             W = lam/(lam-flux*initialXS*3600*1e-24+env.fluence*effectiveXS*3600*1e-24)
-            # Column X: V#*[e(-S#)-e(U#)]
-            if abs(U) < 1e-10 and abs(V) < 1e-10:
-                precision_correction = W * (V-U+(V+U)/2)
-            else:
-                precision_correction = W * (exp(-U)-exp(-V))
+            # Column X: W*(exp(-U)-exp(V)) if U,V > 1e-10 else W*(V-U+(V-U)*(V+U)/2)
+            # [PAK 2024-02-28] Rewrite the exponential difference using expm1()
+            X = W*exp(-V)*expm1(V-U) if U>V else -W*exp(-U)*expm1(U-V)
+            # Column Y: O if "b" else T if "2n" else L*X
+            activity = root*X
+            #print(f"{ai.isotope}=>{ai.daughter} {U=} {V=} {W=} {X=} {activity=}")
 
-            activity = root*precision_correction
             if activity < 0:
                 msg = "activity %g less than zero for %g"%(activity, isotope)
                 raise RuntimeError(msg)
@@ -492,78 +549,151 @@ def init(table, reload=False):
     """
     Add neutron activation levels to each isotope.
     """
-    if 'neutron_activation' in table.properties and not reload:
+    from math import floor
+    # TODO: importlib.reload does not work for iso.neutron_activation attribute
+    if 'neutron_activation' not in table.properties:
+        table.properties.append('neutron_activation')
+    elif not reload:
         return
-    table.properties.append('neutron_activation')
-
-    # Clear the existing activation table
-    for el in table:
-        for iso in el.isotopes:
-            if hasattr(el[iso], 'neutron_activation'):
-                del el[iso].neutron_activation
+    else:
+        # Reloading activation table so clear the existing data.
+        for el in table:
+            for iso in el.isotopes:
+                if hasattr(el[iso], 'neutron_activation'):
+                    del el[iso].neutron_activation
 
+    # We are keeping the table as a simple export of the activation data
+    # from the ncnr health physics excel spreadsheet so that it is easier
+    # to validate that the table contains the same data. Unfortunately some
+    # of the cells involved formulas, which need to be reproduced when loading
+    # in order to match full double precision values.
+    activations = {}
     path = os.path.join(core.get_data_path('.'), 'activation.dat')
-    for row in open(path, 'r'):
-        columns = row.split('\t')
-        if columns[0].strip() in ('', 'xx'):
-            continue
-        columns = [c[1:-1] if c.startswith('"') else c
-                   for c in columns]
-        #print columns
-        for c in INT_COLUMNS:
-            columns[c] = int(columns[c])
-        for c in BOOL_COLUMNS:
-            columns[c] = (columns[c] == 'y')
-        for c in FLOAT_COLUMNS:
-            columns[c] = float(columns[c]) if columns[c].strip() else 0.
-        # clean up comment column
-        columns[-1] = columns[-1].replace('"', '').strip()
-        kw = dict(zip(COLUMN_NAMES, columns))
-        kw['Thalf_str'] = " ".join((kw['_Thalf'], kw['_Thalf_unit']))
+    with open(path, 'r') as fd:
+        for row in fd:
+            #print(row, end='')
+            columns = row.split('\t')
+            if columns[0].strip() in ('', 'xx'):
+                continue
+            columns = [c[1:-1] if c.startswith('"') else c
+                    for c in columns]
+            #print columns
+            for c in INT_COLUMNS:
+                columns[c] = int(columns[c])
+            for c in BOOL_COLUMNS:
+                columns[c] = (columns[c] == 'y')
+            for c in FLOAT_COLUMNS:
+                s = columns[c].strip()
+                columns[c] = float(s) if s else 0.
+            # clean up comment column
+            columns[-1] = columns[-1].replace('"', '').strip()
+            kw = dict(zip(COLUMN_NAMES, columns))
+            iso = (kw['Z'], kw['A'])
+            act = activations.setdefault(iso, [])
 
-        # Strip columns whose names start with underscore
-        kw = dict((k, v) for k, v in kw.items() if not k.startswith('_'))
+            # TODO: use NuBase2020 for halflife
+            # NuBase2020 uses different isomer labels.
+            # Strip the (+, *, s, t) tags
+            # Convert m1/m2 to m/n for
+            #   In-114 In-116 Sb-124 Eu-152 Hf-179 Ir-192
+            # Convert Eu-152m2 to Eu-152r
+            # Convert m to n for
+            #   Sc-46m Ge-73m Kr-83m Pd-107m Pd-109m Ag-110m Ta-182m Ir-194m Pb-204m
+            # Convert m to p for
+            #   Sb-122m Lu-177m
 
-        # Create an Activation record and add it to the isotope
-        iso = table[kw['Z']][kw['A']]
-        activation = getattr(iso, 'neutron_activation', [])
-        activation.append(ActivationResult(**kw))
-        iso.neutron_activation = activation
+            # Recreate Thalf_hrs column using double precision.
+            # Note: spreadsheet is not converting half-life to hours in cell AW1462 (186-W => 188-W)
+            kw['Thalf_hrs'] = float(kw['_Thalf']) * UNITS_TO_HOURS[kw['_Thalf_unit']]
+            #print(f"T1/2 {kw['Thalf_hrs']} +/- {kw['Thalf_hrs_unc']}")
+            # Recreate Thalf_parent by fetching from the new Thalf_hrs
+            # e.g., =IF(OR(AR1408="2n",AR1408="b"),IF(AR1407="b",AW1406,AW1407),"")
+            # This requires that the parent is directly before the 'b' or 'nb'
+            # with its activation list already entered into the isotope.
+            # Note: 150-Nd has 'act' followed by two consecutive 'b' entries.
+            if kw['reaction'] in ('b', '2n'):
+                parent = act[-2] if act[-1].reaction == 'b' else act[-1]
+                kw['Thalf_parent'] = parent.Thalf_hrs
+            else:
+                #assert kw['Thalf_parent'] == 0
+                kw['Thalf_parent'] = None
+            # Half-lives use My, Gy, Ty, Py
+            value, units = float(kw['_Thalf']), kw['_Thalf_unit']
+            if units == 'y':
+                if value >= 1e12:
+                    value, units = value/1e12, 'Ty'
+                elif value >= 1e9:
+                    value, units = value/1e9, 'Gy'
+                elif value >= 200e3: # above 200,000 years use My
+                    value, units = value/1e6, 'My'
+                elif value >= 20e3: # between 20,000 and 200,000 use ky
+                    value, units = value/1e3, 'ky'
+            formatted = f"{value:g} {units}"
+            #if formatted.replace(' ', '') != kw['Thalf_str'].replace(' ', ''):
+            #    print(f"{kw['_index']}: old {kw['Thalf_str']} != {formatted} new")
+            kw['Thalf_str'] = formatted
 
-        # Check abundance values
-        #if abs(iso.abundance - kw['abundance']) > 0.001*kw['abundance']:
-        #    percent = 100*abs(iso.abundance - kw['abundance'])/kw['abundance']
-        #    print "Abundance of", iso, "is", iso.abundance, \
-        #        "but activation.dat has", kw['abundance'], "(%.1f%%)"%percent
+            # Strip columns whose names start with underscore
+            kw = dict((k, v) for k, v in kw.items() if not k.startswith('_'))
+
+            # Create an Activation record and add it to the isotope
+            act.append(ActivationResult(**kw))
+
+            # Check abundance values
+            #if abs(iso.abundance - kw['abundance']) > 0.001*kw['abundance']:
+            #    percent = 100*abs(iso.abundance - kw['abundance'])/kw['abundance']
+            #    print "Abundance of", iso, "is", iso.abundance, \
+            #        "but activation.dat has", kw['abundance'], "(%.1f%%)"%percent
+
+    # Plug the activation products into the table
+    for (Z, A), daughters in activations.items():
+        table[Z][A].neutron_activation = tuple(daughters)
 
 class ActivationResult(object):
     def __init__(self, **kw):
         self.__dict__ = kw
+    def __str__(self):
+        return str(self.__dict__)
 
 
 def demo():  # pragma: nocover
     import sys
-    decay_level = 5e-4
-    formula = sys.argv[1]
-    fluence = 1e5
-    exposure = 10
-    mass = 1
-    if 0: # Make sure all elements compute
-        import periodictable as pt
-        formula = "".join(str(el) for el in pt.elements)[1:]
+    import argparse
+    import periodictable as pt
+
+    parser = argparse.ArgumentParser(description='Process some data with mass, flux, exposure, and decay options.')
+    parser.add_argument('-m', '--mass', type=float, default=1, help='Specify the mass value (default: 1)')
+    parser.add_argument('-f', '--flux', type=float, default=1e8, help='Specify the flux value (default: 1e8)')
+    parser.add_argument('-e', '--exposure', type=float, default=10, help='Specify the exposure value (default: 10)')
+    parser.add_argument('-d', '--decay', type=float, default=5e-4, help='Specify the decay value (default: 5e-4)')
+    parser.add_argument('--cd-ratio', type=float, default=70, help='Specify the Cd ratio value (default: 70)')
+    parser.add_argument('--fast-ratio', type=float, default=50, help='Specify the fast ratio value (default: 50)')
+
+    parser.add_argument('formula', nargs='?', type=str, default=None, help='Specify the formula as a positional argument')
+    args = parser.parse_args()
+
+    formula = args.formula
+    if formula is None:
+        # Make sure all elements compute
+        #formula = "".join(str(el) for el in pt.elements)[1:]
+        formula = build_formula([(1/el.mass, el) for el in pt.elements][1:])
         # Use an enormous mass to force significant activation of rare isotopes
         mass, fluence = 1e15, 1e8
-    env = ActivationEnvironment(fluence=fluence, Cd_ratio=70, fast_ratio=50, location="BT-2")
-    sample = Sample(formula, mass=mass)
+    env = ActivationEnvironment(fluence=args.flux, Cd_ratio=args.cd_ratio, fast_ratio=args.fast_ratio, location="BT-2")
+    sample = Sample(formula, mass=args.mass)
+    abundance = IAEA1987_isotopic_abundance
+    #abundance=NIST2001_isotopic_abundance,
     sample.calculate_activation(
-        env, exposure=exposure, rest_times=(0, 1, 24, 360),
-        abundance=IAEA1987_isotopic_abundance,
-        #abundance=NIST2001_isotopic_abundance,
+        env, exposure=args.exposure, rest_times=(0, 1, 24, 360),
+        abundance=abundance,
+        )
+    decay_time = sample.decay_time(args.decay)
+    print(f"{args.mass} g {formula} for {args.exposure} hours at {args.flux} n/cm^2/s")
+    print(f"Time to decay to {args.decay} uCi is {decay_time} hours.")
+    sample.calculate_activation(
+        env, exposure=args.exposure, rest_times=(0, 1, 24, 360, decay_time),
+        abundance=abundance,
         )
-    print("%gg %s for %g hours at %g n/cm^2/s"
-          % (mass, formula, exposure, fluence))
-    print("Time to decay to %g uCi is %g hours."
-          % (decay_level, sample.decay_time(decay_level)))
     sample.show_table(cutoff=0.0)
 
     ## Print a table of flux vs. activity so we can debug the
